//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_uail_gui.hpp"
#include "ez_wave.hpp"

//############################################################################
ez_uail_app::ez_uail_app()
            : ez_uail                 (          )
            , black_threshold         (0         )
            , wave_sample_rate        (48000     ) // Hz
            , lit_dwell_overhang      (2         )
            , output_bmp_size         (1000      ) // pxls square
            , lit_distance_delta_max  (600       ) // points
            , blank_distance_delta_max(900       ) // points
            , longest_dwell_microsec  (333       )
            , insignificant_distance  (60        ) // points
            , zero_angle_tolerance    (one_degree) // radians
            , recolor_span_factor     (2.0       )
            , rendered_arc_angle      (6.0       ) // degrees
            , frames_per_second       (16        ) // 0 == every frame is shown exactly once
            , still_frame_duration    (10.0      ) // seconds
            , auto_scale_dxf          (true      )
            , maintain_dxf_origin     (false     )
            , black_dwell_points      (true      )
            , auto_flatten_z          (true      )
            , expensive_minimize      (false     )
            , invert_wave_output      (false     )
            , auto_optimize           (true      )
            , no_black_match          (false     )
            , dumb_file_format        (false     )
{
    wave_offsets    = new int[8];
    wave_offsets[0] = 0;
    wave_offsets[1] = 0;
    wave_offsets[2] = 0;
    wave_offsets[3] = 0;
    wave_offsets[4] = 0;
    wave_offsets[5] = 0;
    wave_offsets[6] = 0;
    wave_offsets[7] = 0;
}

//############################################################################
ez_uail_app::~ez_uail_app()
{
    delete wave_offsets;
}

//############################################################################
void ez_uail_app::apply_wave_offsets(string file_name)
{
    int                      i                         ,
                             span                      ,
                             last                      ,
                             offset_in_max             ,
                             offset_in_min             ,
                             offset_applied_max        ,
                             offset_applied_min        ,
                            *offset_applied     = NULL ;
    bool                     inverted           = false;
    ifstream                 wave_ifstream             ;
    ofstream                 wave_ofstream             ;
    ez_wave_header           wave_in_header            ,
                             wave_out_header           ;                         
    ez_wave_sample           wave_sample(6)            ;
    ez_wave_sample_pointer   roll_over                 ;
    ez_wave_sample_pointer  *sample_window             ;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to remane      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.from_ifstream_wave(wave_ifstream))
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels != 6)
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s is not a 6 channel wave"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_out_header = wave_in_header;
    wave_out_header.ez_wave_mode |= EZ_WAVE_OFFSETS;
    //------------------------------------------------------------------------
    offset_applied  = new int[6];
    //------------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        offset_applied[i] = wave_in_header.offset[i] - wave_offsets[i];
    //------------------------------------------------------------------------
    offset_in_min      = wave_offsets  [0];
    offset_in_max      = wave_offsets  [0];
    offset_applied_min = offset_applied[0];
    offset_applied_max = offset_applied[0];
    //------------------------------------------------------------------------
    for(i = 1; i < 6; i++)
    {
        if(wave_offsets  [i] > offset_in_max     ) offset_in_max      = wave_offsets  [i];
        if(wave_offsets  [i] < offset_in_min     ) offset_in_min      = wave_offsets  [i];
        if(offset_applied[i] > offset_applied_max) offset_applied_max = offset_applied[i];
        if(offset_applied[i] < offset_applied_min) offset_applied_min = offset_applied[i];
    }    
    //------------------------------------------------------------------------
    span = (offset_applied_max - offset_applied_min) + 1;
    last = span - 1;
    //------------------------------------------------------------------------
    for(i = 0; i < 6; i++)
    {
        wave_out_header.offset[i] = -(offset_in_max - wave_offsets[i]);
        offset_applied[i] = offset_applied_max - offset_applied[i];
    }
    //------------------------------------------------------------------------
    sample_window = new ez_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new ez_wave_sample(6); // span element array of (6 channel ez_wave_samples)
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // take up the space!
    //------------------------------------------------------------------------
    while(true)
    {
        if(last) // span > 1
        {
            roll_over = sample_window[0];
            for(i = 0; i < last; i++)
                sample_window[i] = sample_window[i + 1];
            sample_window[last] = roll_over;
        }

        if(!sample_window[last]->from_ifstream_wave(wave_ifstream))
            break;
            
        for(i = 0; i < 6; i++) // each channel
            wave_sample.channel[i] = sample_window[offset_applied[i]]->channel[i];
        wave_sample.to_ofstream_wave(wave_ofstream, wave_out_header); // write it to disk
    }
    delete sample_window;
    delete offset_applied;
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream);
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    return;    
}

//############################################################################
void ez_uail_app::overwirte_wave_offsets(string file_name)
{
    int                  i                ;
    ifstream             wave_ifstream    ;
    ofstream             wave_ofstream    ;
    ez_wave_header       wave_in_header   ,
                         wave_out_header  ;
    ez_wave_sample       wave_sample      ;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to remane      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.from_ifstream_wave(wave_ifstream))
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_out_header = wave_in_header;
    //------------------------------------------------------------------------
    for(i = 0; i < wave_in_header.num_channels; i++)
        wave_out_header.offset[i] = wave_offsets[i];
    //------------------------------------------------------------------------
    wave_out_header.ez_wave_mode |= EZ_WAVE_OFFSETS;
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // take up the space!
    //------------------------------------------------------------------------
    while(wave_sample.from_ifstream_wave(wave_ifstream))
        wave_sample.to_ofstream_wave(wave_ofstream, wave_out_header);
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream);
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());    
    //------------------------------------------------------------------------
    return;    
}

//############################################################################
void ez_uail_app::invert_wave(string file_name)
{
    int                  i                ;
    ifstream             wave_ifstream    ;
    ofstream             wave_ofstream    ;
    ez_wave_header       wave_in_header   ,
                         wave_out_header  ;
    ez_wave_sample       wave_sample      ;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to remane      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.from_ifstream_wave(wave_ifstream))
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_out_header = wave_in_header;
    //------------------------------------------------------------------------
    if(wave_in_header.ez_wave_mode & EZ_WAVE_POSITIVE)
        wave_out_header.ez_wave_mode &= ~EZ_WAVE_POSITIVE;
    else
        wave_out_header.ez_wave_mode |= EZ_WAVE_POSITIVE;
    //------------------------------------------------------------------------
    wave_out_header.num_samples = 0;
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // take up the space!
    //------------------------------------------------------------------------
    while(wave_sample.from_ifstream_wave_inverted(wave_ifstream))
        wave_sample.to_ofstream_wave(wave_ofstream, wave_out_header);
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream);
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());    
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::split_wave_xyrgbi(string file_name)
{
    int                  i                 ;
    ifstream             wave_ifstream     ;
    ofstream             wave_ofstream_xy  ,
                         wave_ofstream_r   ,
                         wave_ofstream_g   ,
                         wave_ofstream_b   ,
                         wave_ofstream_i   ;
    ez_wave_header       wave_in_header    ,
                         wave_out_header_xy,
                         wave_out_header_r ,
                         wave_out_header_g ,
                         wave_out_header_b ,
                         wave_out_header_i ;
    ez_wave_sample       wave_sample       ,
                         stereo_sample(2)  ,
                         mono_sample  (1)  ;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.from_ifstream_wave(wave_ifstream))
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels < 6)
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s is not a 6 channel wave"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    file_name = file_name.substr(0, file_name.size() - 4);
    //------------------------------------------------------------------------
    wave_ofstream_xy.open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name + "_xy.wav").c_str(), ios::out | ios::binary);
    wave_ofstream_r .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name + "_r.wav" ).c_str(), ios::out | ios::binary);
    wave_ofstream_g .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name + "_g.wav" ).c_str(), ios::out | ios::binary);
    wave_ofstream_b .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name + "_b.wav" ).c_str(), ios::out | ios::binary);    
    wave_ofstream_i .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name + "_i.wav" ).c_str(), ios::out | ios::binary);
    //------------------------------------------------------------------------
    if(    !wave_ofstream_xy.is_open()
        || !wave_ofstream_r .is_open()
        || !wave_ofstream_g .is_open()
        || !wave_ofstream_b .is_open()
        || !wave_ofstream_i .is_open()
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_SPLIT_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_out_header_xy = wave_in_header;
    wave_out_header_r  = wave_in_header;
    wave_out_header_g  = wave_in_header;
    wave_out_header_b  = wave_in_header;
    wave_out_header_i  = wave_in_header;
    //------------------------------------------------------------------------
    wave_out_header_xy.num_samples = 0;
    wave_out_header_r .num_samples = 0;
    wave_out_header_g .num_samples = 0;
    wave_out_header_b .num_samples = 0;
    wave_out_header_i .num_samples = 0;
    //------------------------------------------------------------------------
    wave_out_header_xy.num_channels = 2; // stereo    
    wave_out_header_r .num_channels = 1; // mono
    wave_out_header_g .num_channels = 1; // mono
    wave_out_header_b .num_channels = 1; // mono
    wave_out_header_i .num_channels = 1; // mono
    //------------------------------------------------------------------------
    while(wave_sample.from_ifstream_wave(wave_ifstream))
    {
        stereo_sample.channel[0] = wave_sample.channel[0];
        stereo_sample.channel[1] = wave_sample.channel[1];
        stereo_sample.to_ofstream_wave(wave_ofstream_xy, wave_out_header_xy);
        mono_sample.channel[0]   = wave_sample.channel[2];
        mono_sample.to_ofstream_wave(wave_ofstream_r, wave_out_header_r);
        mono_sample.channel[0]   = wave_sample.channel[3];
        mono_sample.to_ofstream_wave(wave_ofstream_g, wave_out_header_g);
        mono_sample.channel[0]   = wave_sample.channel[4];
        mono_sample.to_ofstream_wave(wave_ofstream_b, wave_out_header_b);
        mono_sample.channel[0]   = wave_sample.channel[5];
        mono_sample.to_ofstream_wave(wave_ofstream_i, wave_out_header_i);
    }
    //------------------------------------------------------------------------
    wave_out_header_xy.to_ofstream_wave(wave_ofstream_xy, false);
    wave_out_header_r .to_ofstream_wave(wave_ofstream_r , false);
    wave_out_header_g .to_ofstream_wave(wave_ofstream_g , false);
    wave_out_header_b .to_ofstream_wave(wave_ofstream_b , false);
    wave_out_header_i .to_ofstream_wave(wave_ofstream_i , false);
    //------------------------------------------------------------------------
    wave_ifstream   .close();
    wave_ofstream_xy.close();
    wave_ofstream_r .close();
    wave_ofstream_g .close();
    wave_ofstream_b .close();
    wave_ofstream_i .close();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::join_xyrgbi_waves(string file_name_xy,
                                    string file_name_r,
                                    string file_name_g,
                                    string file_name_b,
                                    string file_name_i,
                                    string file_name_out
                               )
{
    ifstream       in_xy,
                   in_r,
                   in_g,
                   in_b,
                   in_i;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    if(    file_name_xy.size() < 4
        || file_name_xy.substr(file_name_xy.size() - 4) != ".wav"
      )
        file_name_xy += ".wav";        
    if(    file_name_r.size() < 4
        || file_name_r.substr(file_name_r.size() - 4) != ".wav"
      )
        file_name_r += ".wav";        
    if(    file_name_g.size() < 4
        || file_name_g.substr(file_name_g.size() - 4) != ".wav"
      )
        file_name_g += ".wav";        
    if(    file_name_b.size() < 4
        || file_name_b.substr(file_name_b.size() - 4) != ".wav"
      )
        file_name_b += ".wav";        
    if(    file_name_i.size() < 4
        || file_name_i.substr(file_name_i.size() - 4) != ".wav"
      )
        file_name_i += ".wav";
    if(    file_name_out.size() < 4
        || file_name_out.substr(file_name_out.size() - 4) != ".wav"
      )
        file_name_out += ".wav";
    //------------------------------------------------------------------------
    in_xy.open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name_xy).c_str(), ios::in | ios::binary);
    in_r .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name_r ).c_str(), ios::in | ios::binary);
    in_g .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name_g ).c_str(), ios::in | ios::binary);
    in_b .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name_b ).c_str(), ios::in | ios::binary);
    in_i .open((EZ_UAIL_SPLIT_WAVE_SHARE + file_name_i ).c_str(), ios::in | ios::binary);
    //------------------------------------------------------------------------
    if(    in_xy.is_open()
        && in_r .is_open()
        && in_g .is_open()
        && in_b .is_open()
        && in_i .is_open()
      )
    {
        ofstream       out;        
        ez_wave_header wave_in_header,
                       wave_out_header;
        ez_wave_sample sample_xyrgbi(6),
                       sample_xy    (2),
                       sample_r     (1),
                       sample_g     (1),
                       sample_b     (1),
                       sample_i     (1);
        //--------------------------------------------------------------------        
        out.open((EZ_UAIL_WAVE_SHARE + file_name_out).c_str(), ios::out | ios::binary);
        //--------------------------------------------------------------------        
        wave_out_header.ez_wave_mode |= EZ_WAVE_OFFSETS;
        //--------------------------------------------------------------------
        wave_in_header.from_ifstream_wave(in_xy);
        wave_in_header.from_ifstream_wave(in_r);
        wave_in_header.from_ifstream_wave(in_g);
        wave_in_header.from_ifstream_wave(in_b);
        wave_in_header.from_ifstream_wave(in_i);
        //--------------------------------------------------------------------        
        wave_out_header.to_ofstream_wave(out); // take up the space!
        //--------------------------------------------------------------------
        while(    sample_xy.from_ifstream_wave(in_xy)
               && sample_r .from_ifstream_wave(in_r)
               && sample_g .from_ifstream_wave(in_g)
               && sample_b .from_ifstream_wave(in_b)
               && sample_i .from_ifstream_wave(in_i)
             )
        {
            sample_xyrgbi =   sample_xy
                            + sample_r 
                            + sample_g 
                            + sample_b 
                            + sample_i; // isn't that neat!
                            
            sample_xyrgbi.to_ofstream_wave(out, wave_out_header);
        }
        //--------------------------------------------------------------------
        wave_out_header.to_ofstream_wave(out);
        //--------------------------------------------------------------------
        in_xy.close();
        in_r .close();
        in_g .close();
        in_b .close();
        in_i .close();
        out  .close();
    }
    //------------------------------------------------------------------------
    else
    {
        AIL->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(AIL->bg_color.r,
                                   AIL->bg_color.g,
                                   AIL->bg_color.b
                                  )
                    , 0
                    , AIL->font_size_factor
                    , "[any key] at least one of those files failed to open"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_UAIL_FILE_OPEN_FAILED;
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::call_frame_effect(string name)
{
    int i;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    for(i = 0; i < (int)frame_effect_names.size(); i++)
        if(frame_effect_names[i] == name)
            break;
    //------------------------------------------------------------------------            
    if(i < (int)frame_effect_names.size())
        frame_effects[i](&(current_frame()));
    //------------------------------------------------------------------------
    else
    {
        AIL->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(AIL->bg_color.r,
                                   AIL->bg_color.g,
                                   AIL->bg_color.b
                                  )
                    , 0
                    , AIL->font_size_factor
                    , "[any key] no frame effect of that name found"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_EFFECT_NAME_NOT_FOUND;
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::call_frame_set_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        if(frame_set_effect_names[i] == name)
            break; 
    if(i < (int)frame_set_effect_names.size())
        frame_set_effects[i](this);
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
